unit hiHTTPClient;

interface

{$I share.inc}

uses
  Kol, Share, WinInet, WinHTTP, Windows, Debug, URI, MIMETypes,
  hiCookieManager, hiCharset;

type

  ThiHTTPClient = class(TDebug)
    private
      FURLParser: TURI;
      FAbort: Boolean;
      
      //   
      FdtURL: string;
      FdtFileName: string;
      FdtHeaders: string;
      FdtHTTPUser: string;
      FdtHTTPPass: string;
      FdtMethod: WideString;
      FdtPostData: string;
      FdtProxy: string;
      FdtProxyUser: string;
      FdtProxyPass: string;
      FdtReferer: string;
      FdtUploadFile: string;
      FdtUploadStream: PStream;
      // ------------------- //
      
      //    (   )
      FStatusHeader: Integer;
      FContLenHeader: Int64;
      FContTypeHeader: string;
      FResponseHeaders: string;
      // ------------------- //
      
      //   
      FHSession: HINTERNET;
      FHConnect: HINTERNET;
      FHRequest: HINTERNET;
      
      FRedirected: Boolean;
      FNeedBody: Boolean;
      FBodyLength: DWord;
      FReqBody: string;
      FReqHeaders: string;
      FULStream: PStream;
      FDLStream: PStream;
      FReleaseULStream: Boolean;
      FReleaseDLStream: Boolean;
      // ------------------- //
      
      procedure InternalDownload(var _Data: TData; Mode: Integer);
      procedure ReadPoints(var _Data: TData; Mode: Integer);
      procedure ReadUploadFile(var _Data: TData);
      function PrepareSave(var _Data: TData; Mode: Integer): DWord;
      function PrepareUpload: DWord;
      procedure PrepareRequest;
      function StartRequest: DWord;
      function SendRequest: DWord;
      function SendBody: DWord;
      function SaveResponse: DWord;
      procedure CancelRequest;
      procedure Cleanup(IsFailed: Boolean);
      
    public
      _prop_Boundary: string;
      _prop_CustomMethod: string;
      _prop_FileField: string;
      _prop_FileName: string;
      _prop_FileMIME: string;
      _prop_FNOverride: string;
      _prop_Headers: string;
      _prop_PostData: string;
      _prop_Proxy: string;
      _prop_ProxyUser: string;
      _prop_ProxyPass: string;
      _prop_Referer: string;
      _prop_UploadFile: string;
      _prop_URL: string;
      _prop_UserAgent: string;
      _prop_HTTPUser: string;
      _prop_HTTPPass: string;
      _prop_AuthCheck: Byte;
      _prop_AutoRedirect: Byte;
      _prop_ConnectTimeout: Integer;
      _prop_EncType: Byte;
      _prop_Method: Byte;
      _prop_RecvTimeout: Integer;
      _prop_ResumeDL: Byte;
      _prop_SendTimeout: Integer;
      _prop_SSLProtocols: Byte;
      _prop_URLCodePage: Byte;
      _prop_UseSystemProxy: Byte;
      _prop_IgnoreInvalidCN: Byte;
      _prop_IgnoreInvalidDate: Byte;
      _prop_IgnoreUnknownCA: Byte;
      _prop_IgnoreWeakSignature: Byte;
      _prop_IgnoreWrongUsage: Byte;
      _prop_CookieManager: ThiCookieManager;

      _data_FileName: THI_Event;
      _data_HTTPPass: THI_Event;
      _data_HTTPUser: THI_Event;
      _data_Headers: THI_Event;
      _data_PostData: THI_Event;
      _data_Proxy: THI_Event;
      _data_ProxyPass: THI_Event;
      _data_ProxyUser: THI_Event;
      _data_Referer: THI_Event;
      _data_SaveTo: THI_Event;
      _data_Stream: THI_Event;
      _data_URL: THI_Event;
      _data_UploadFile: THI_Event;
      
      _event_onDownloadProgress: THI_Event;
      _event_onError: THI_Event;
      _event_onFinish: THI_Event;
      _event_onHeadersAvail: THI_Event;
      _event_onRedirect: THI_Event;
      _event_onStateChange: THI_Event;
      _event_onUploadProgress: THI_Event;
      
      constructor Create;
      destructor Destroy; override;
      
      procedure _work_doAbort(var _Data: TData; Index: Word);
      procedure _work_doBoundary(var _Data: TData; Index: Word);
      procedure _work_doCustomMethod(var _Data: TData; Index: Word);
      procedure _work_doEncType(var _Data: TData; Index: Word);
      procedure _work_doExecute(var _Data: TData; Index: Word);
      procedure _work_doFileField(var _Data: TData; Index: Word);
      procedure _work_doFileMIME(var _Data: TData; Index: Word);
      procedure _work_doFNOverride(var _Data: TData; Index: Word);
      procedure _work_doLoadFile(var _Data: TData; Index: Word);
      procedure _work_doLoadStream(var _Data: TData; Index: Word);
      procedure _work_doLoadString(var _Data: TData; Index: Word);
      procedure _work_doResumeDL(var _Data: TData; Index: Word);
      procedure _work_doUseSystemProxy(var _Data: TData; Index: Word);
      
      procedure _var_Status(var _Data: TData; Index: Word);
      procedure _var_ContentLength(var _Data: TData; Index: Word);
      procedure _var_ContentType(var _Data: TData; Index: Word);
      procedure _var_CurrentURL(var _Data: TData; Index: Word);
      procedure _var_RawHeaders(var _Data: TData; Index: Word);
      procedure _var_StatusText(var _Data: TData; Index: Word);
  end;

implementation


const
  RequestMethods: array [0..2] of WideString = ('GET', 'POST', 'HEAD');
  DefaultBoundary = '----HiHtTpClIeNt';
  DefaultUploadFN = 'file.bin';
  
  AUTH_CHECK_DISABLED = 0;
  AUTH_CHECK_PROXY = 1;
  AUTH_CHECK_ALL = 2;
  
  METHOD_GET = 0;
  METHOD_POST = 1;
  METHOD_HEAD = 2;
  
  ENC_TYPE_MULTIPART = 0;
  ENC_TYPE_URL_ENC = 1;
  ENC_TYPE_NONE = 2;
  
  PROMPT_FOR_SAVE = -1;
  SAVE_TO_STRING = 1;
  SAVE_TO_FILE = 2;
  SAVE_TO_STREAM = 3;


//  Replace = True     
//  Replace = False       
procedure AddHeadersToRequest(RequestH: HINTERNET; const Headers: string; Replace: Boolean = True);
var
  F: DWord;
begin
  if Headers = '' then Exit;
  
  F := WINHTTP_ADDREQ_FLAG_ADD;
  if Replace then F := F or WINHTTP_ADDREQ_FLAG_REPLACE;
  
  WinHttpAddRequestHeaders(
    RequestH,
    PWideChar(WideString(Headers)),
    Length(Headers),
    F
  );
end;

function ChooseAuthScheme(RequestH: HINTERNET): DWord;
var
  SuppSchemes, FirstScheme, Target: DWord;
begin
  Result := WINHTTP_AUTH_SCHEME_DIGEST;
  if WinHttpQueryAuthSchemes(RequestH, SuppSchemes, FirstScheme, Target) then
  begin
    if SuppSchemes and WINHTTP_AUTH_SCHEME_DIGEST <> 0 then
      Result := WINHTTP_AUTH_SCHEME_DIGEST
    else
      Result := WINHTTP_AUTH_SCHEME_BASIC;
  end;
end;

function GetHTTPOptionDWord(Handle: HINTERNET; Option: DWord): DWord;
var
  L: DWord;
begin
  L := SizeOf(DWord);
  WinHttpQueryOption(Handle, Option, @Result, L);
end;

function GetHTTPOptionString(Handle: HINTERNET; Option: DWord): string;
var
  Buffer: WideString;
  BufLen: DWord;
  Rslt: Boolean;
label
  start;
const
  OPTION_BUF_SIZE = 250;
begin
  Result := '';
  if Handle = nil then Exit;
  
  BufLen := OPTION_BUF_SIZE;
  SetLength(Buffer, OPTION_BUF_SIZE);
  
start:
  Rslt := WinHttpQueryOption(Handle, Option, PWideChar(Buffer), BufLen);

  if (Rslt = False) and (GetLastError() = ERROR_INSUFFICIENT_BUFFER) then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    goto start;
  end;

  if Rslt then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    Result := Buffer;
  end;

end;

function ReadResponseHeader(Handle: HINTERNET; HeaderIdx: DWord; PosIdx: PDWord = nil): string; overload;
var
  Buffer: WideString;
  BufLen, PrevIdx: DWord;
  Rslt: Boolean;
label
  start;
const
  HEADER_BUF_SIZE = 250;
begin
  Result := '';
  PrevIdx := 0;
  if Handle = nil then Exit;
  
  BufLen := HEADER_BUF_SIZE;
  SetLength(Buffer, HEADER_BUF_SIZE);
  
start:
  if PosIdx <> nil then PrevIdx := PosIdx^;
  Rslt := WinHttpQueryHeaders(
             Handle,
             HeaderIdx,
             nil,
             PWideChar(Buffer),
             BufLen,
             PosIdx
           );

  if (Rslt = False) and (GetLastError() = ERROR_INSUFFICIENT_BUFFER) then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    if PosIdx <> nil then PosIdx^ := PrevIdx;
    goto start;
  end;


  if Rslt then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    Result := Buffer;
  end;
    
end;

function ReadResponseHeader(Handle: HINTERNET; HeaderName: string; PosIdx: PDWord = nil): string; overload;
var
  Buffer: WideString;
  BufLen, PrevIdx: DWord;
  Rslt: Boolean;
label
  start;
const
  HEADER_BUF_SIZE = 250;
begin
  Result := '';
  PrevIdx := 0;
  if Handle = nil then Exit;
  
  BufLen := HEADER_BUF_SIZE;
  SetLength(Buffer, HEADER_BUF_SIZE);

start:
  if PosIdx <> nil then PrevIdx := PosIdx^;
  Rslt := WinHttpQueryHeaders(
             Handle,
             WINHTTP_QUERY_CUSTOM,
             PWideChar(WideString(HeaderName)),
             PWideChar(Buffer),
             BufLen,
             PosIdx //WINHTTP_NO_HEADER_INDEX
           );

  if (Rslt = False) and (GetLastError() = ERROR_INSUFFICIENT_BUFFER) then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    if PosIdx <> nil then PosIdx^ := PrevIdx;
    goto start;
  end;

  if Rslt then
  begin
    SetLength(Buffer, BufLen div SizeOf(WideChar));
    Result := Buffer;
  end;
    
end;

function ReadResponseCookies(Handle: HINTERNET): string;
var
  I: DWord;
  Lst: PStrList;
  S: string;
begin
  Result := '';
  I := 0;
  Lst := NewStrList;
  
  while True do
  begin
    S := ReadResponseHeader(Handle, WINHTTP_QUERY_SET_COOKIE, @I);
    if S <> '' then Lst.Add(S)
    else Break;
  end;
  if Lst.Count > 0 then Result := Lst.Text;
  Lst.Free;
end;

procedure SetAuthCredentials(RequestH: HINTERNET; const User, Pass: string; Target, Scheme: DWord);
begin
  if (User = '') or (Pass = '') then Exit;

  WinHttpSetCredentials(
    RequestH,
    Target,
    Scheme,
    PWideChar(WideString(User)),
    PWideChar(WideString(Pass)),
    nil
  );
end;

procedure SetHTTPOptionDWord(RequestH: HINTERNET; Option, Value: DWord);
begin
  WinHttpSetOption(RequestH, Option, @Value, SizeOf(DWord));
end;








const
  CallbackFlags = 
    WINHTTP_CALLBACK_FLAG_RESOLVE_NAME or
    WINHTTP_CALLBACK_FLAG_CONNECT_TO_SERVER or
    WINHTTP_CALLBACK_FLAG_SEND_REQUEST or
    WINHTTP_CALLBACK_FLAG_RECEIVE_RESPONSE or
    WINHTTP_CALLBACK_FLAG_REDIRECT;

procedure StatusCallback(Handle: HINTERNET; dwContext: Pointer; dwInternetStatus: DWord; lpvStatusInformation: Pointer; dwStatusInformationLength: DWord); stdcall;
var
  U: string;
  UrlP: PWideChar;
  Client: ThiHTTPClient;
begin

  Client := dwContext;
  if dwInternetStatus = WINHTTP_CALLBACK_STATUS_REDIRECT then
  begin
    UrlP := lpvStatusInformation;
    U := UrlP;
    Client.FRedirected := True;
      
    _hi_OnEvent(Client._event_onRedirect, U);
    

    //  ,    
    AddHeadersToRequest(Handle, 'Cookie: ');
    
    if Client._prop_CookieManager <> nil then
    begin
      //       
      Client._prop_CookieManager.AddCookies(GetHTTPOptionString(Handle, WINHTTP_OPTION_URL), ReadResponseCookies(Handle));

      //     
      U := Client._prop_CookieManager.GetCookies(U);
      if U <> '' then AddHeadersToRequest(Handle, U);
    end;
    Exit;
  end;
  
  _hi_OnEvent(Client._event_onStateChange, Trunc(ln(dwInternetStatus) / ln(2)));
end;

//    
function StreamToString(Stream: PStream): string;
var
  R: Integer;
begin
  if Stream <> nil then
  begin
    SetLength(Result, Stream.Size);
    Stream.Position := 0;
    R := Stream.Read(Result[1], Stream.Size);
    if R <> Length(Result) then SetLength(Result, R);
  end;
end;

//     4 
function StreamGetSize(Stream: PStream): Int64;
begin
  Result := 0;
  if Stream = nil then Exit;
  
  if Stream.Memory <> nil then
    Result := Stream.Size
  else
  begin
    I64(Result).Lo := GetFileSize(Stream.Handle, @I64(Result).Hi);
    if (I64(Result).Lo = INVALID_FILE_SIZE { $FFFFFFFF }) and (GetLastError <> NO_ERROR)
    then
      Result := 0;
  end;
end;

function StreamSeekEOF(Stream: PStream): Int64;
begin
  Result := 0;
  if Stream = nil then Exit;
  
  if Stream.Memory <> nil then
    Result := Stream.Seek(0, spEnd)
  else
  begin
    I64(Result).Lo := SetFilePointer(Stream.Handle, 0, @I64(Result).Hi, FILE_END);
    if (I64(Result).Lo = $FFFFFFFF {INVALID_SET_FILE_POINTER}) and (GetLastError <> NO_ERROR)
    then
      Result := 0;
  end;
end;


// ========== ThiHTTPClient ========== //


constructor ThiHTTPClient.Create;
begin
  inherited;
  FURLParser := TURI.Create;
end;

destructor ThiHTTPClient.Destroy;
begin
  CancelRequest;
  FURLParser.Free;
  inherited;
end;


procedure ThiHTTPClient.InternalDownload(var _Data: TData; Mode: Integer);
label
  finish;
var
  dwError: DWord;
begin
  if FHSession <> nil then Exit; //   
  
  FAbort := False;
  FStatusHeader := 0;
  FContLenHeader := 0;
  FContTypeHeader := '';
  FResponseHeaders := '';
  
  // 1)      
  ReadPoints(_Data, Mode);
  
  // 2)  ,    
  if Mode <> PROMPT_FOR_SAVE then // Mode = -1 - PrepareSave()     
  begin
    dwError := PrepareSave(_Data, Mode);
    if dwError <> NO_ERROR then goto finish;
  end;
  
  // 3)  /  
  dwError := PrepareUpload();
  if dwError <> NO_ERROR then goto finish;
  
  // 4)     
  PrepareRequest();
  
  // 5)     API
  dwError := StartRequest();
  if dwError <> NO_ERROR then goto finish;
  
  // 6)      
  dwError := SendRequest();
  if dwError <> NO_ERROR then goto finish;
  
  // 7)   
  _hi_OnEvent(_event_onHeadersAvail, FStatusHeader);
  if FAbort then
  begin
    dwError := ERROR_OPERATION_ABORTED;
    goto finish;
  end;
  
  // 8)         PROMPT_FOR_SAVE
  if Mode = PROMPT_FOR_SAVE then
  begin
    Mode := ReadInteger(_Data, _data_SaveTo, 0);
    //   .  SAVE_TO_STREAM  Stream   PrepareSave()
    if Mode = SAVE_TO_FILE then
      FdtFileName := ReadString(_Data, _data_FileName, _prop_FileName);
    
    dwError := PrepareSave(_Data, Mode);
    if dwError <> NO_ERROR then goto finish;
  end;  
  
  // 9)        
  dwError := SaveResponse();
  if dwError <> NO_ERROR then goto finish;
  
finish:

  // 10)  
  if dwError = NO_ERROR then
  begin
    if Mode = SAVE_TO_STRING then
      _hi_CreateEvent(_Data, @_event_onFinish, StreamToString(FDLStream))
    else
      _hi_CreateEvent(_Data, @_event_onFinish)
  end
  else
    if not FAbort then
      _hi_CreateEvent(_Data, @_event_onError, Integer(dwError));
  
  // 11)   
  Cleanup(dwError <> NO_ERROR);
end;


procedure ThiHTTPClient.ReadPoints(var _Data: TData; Mode: Integer);
begin
  FdtURL := ReadString(_Data, _data_URL, _prop_URL);
  
  //      doLoadFile
  if Mode = SAVE_TO_FILE then
    FdtFileName := ReadString(_Data, _data_FileName, _prop_FileName);
  
  FdtHeaders := ReadString(_Data, _data_Headers, _prop_Headers);
  FdtHTTPUser := ReadString(_Data, _data_HTTPUser, _prop_HTTPUser);
  FdtHTTPPass := ReadString(_Data, _data_HTTPPass, _prop_HTTPPass);
  
  if _prop_CustomMethod <> '' then
  begin
    FdtMethod := _prop_CustomMethod;
    FNeedBody := True;
  end
  else
  begin
    FdtMethod := RequestMethods[_prop_Method];
    FNeedBody := (_prop_Method = METHOD_POST);
  end;
  if FNeedBody then FdtPostData := ReadString(_Data, _data_PostData, _prop_PostData);
  
  if _prop_UseSystemProxy = 0 then FdtProxy := ReadString(_Data, _data_Proxy, _prop_Proxy);
  FdtProxyUser := ReadString(_Data, _data_ProxyUser, _prop_ProxyUser);
  FdtProxyPass := ReadString(_Data, _data_ProxyPass, _prop_ProxyPass);
  FdtReferer := ReadString(_Data, _data_Referer, _prop_Referer);
  

  if FNeedBody and (_prop_EncType = ENC_TYPE_MULTIPART) then
    ReadUploadFile(_Data);
end;


procedure ThiHTTPClient.ReadUploadFile(var _Data: TData);
var
  D, Def: TData;
begin
  dtString(Def, _prop_UploadFile);
  D := ReadData(_Data, _data_UploadFile, @Def);
  
  FdtUploadStream := ToStream(D);
  if FdtUploadStream = nil then
    FdtUploadFile := ToString(D);
end;


function ThiHTTPClient.PrepareSave(var _Data: TData; Mode: Integer): DWord;
var
  Flags: DWord;
begin
  Result := NO_ERROR;
  
  //if _prop_Method = METHOD_HEAD then Exit; //      _prop_CustomMethod
  
  case Mode of
    SAVE_TO_STRING: // String
      begin
        FDLStream := NewMemoryStream;
        FReleaseDLStream := True;
      end;
    SAVE_TO_FILE: // File
      begin
        Flags := ofOpenWrite or ofShareDenyNone;
        // FHSession <> nil - , PrepareSave()    SaveTo,  
        if (_prop_ResumeDL <> 0) and (FHSession = nil) then
          Flags := Flags or ofOpenAlways
        else
          Flags := Flags or ofCreateAlways; //TODO:     (,     ?)
          
        FDLStream := NewFileStream(FdtFileName, Flags);
        
        if FDLStream.Handle = INVALID_HANDLE_VALUE then
        begin
          Result := GetLastError;
          Free_And_Nil(FDLStream);
          Exit;
        end;
        FReleaseDLStream := True;
      end;    
    SAVE_TO_STREAM: // Stream
      begin
        FDLStream := ReadStream(_Data, _data_Stream);
        if FDLStream = nil then
        begin
          Result := ERROR_FILE_NOT_FOUND;
          Exit;
        end;
        FReleaseDLStream := False;
      end;
  end;
end;


function ThiHTTPClient.PrepareUpload: DWord;
begin
  Result := NO_ERROR;
  if (not FNeedBody) or (_prop_FileField = '') then Exit;
  
  if FdtUploadStream <> nil then
  begin
    FULStream := FdtUploadStream;
  end
  else
  begin
    if FdtUploadFile <> '' then
    begin
      FULStream := NewFileStream(FdtUploadFile, ofOpenRead or ofOpenExisting or ofShareDenyNone);
      //    
      if FULStream.Handle = INVALID_HANDLE_VALUE then
      begin
        Result := GetLastError;
        Free_And_Nil(FULStream);
      end
      else
      begin
        FReleaseULStream := True;
      end;
    end;
  end;
end;


procedure ThiHTTPClient.PrepareRequest; //  ,  
var
  Key, Val: string;
  Body, PostDt: PStrList;
  I: Integer;
  L: DWord;
  L64: Int64;
begin

  //    
  if (_prop_ResumeDL <> 0) and (FDLStream <> nil) then
  begin
    //        
    //L := FDLStream.Seek(0, spEnd); // ,    4 
    L64 := StreamSeekEOF(FDLStream);
    if L64 > 0 then
      FReqHeaders := FReqHeaders + 'Range: bytes=' + Double2Str(L64) + '-'#13#10;
  end;
  
  
  if FNeedBody then
  begin

    if _prop_EncType = ENC_TYPE_MULTIPART then
    begin
      if _prop_Boundary = '' then _prop_Boundary := DefaultBoundary;
    
      //  POSTData
      Body := NewStrList;
      if FdtPostData <> '' then
      begin
        PostDt := NewStrList;
        PostDt.Text := FdtPostData;
        for I := 0 to PostDt.Count - 1 do
        begin
          Val := PostDt.Items[I];
          Key := Parse(Val, '=');
          if Key <> '' then
          begin
            Body.Add('--' + _prop_Boundary);
            Body.Add('Content-Disposition: form-data; name="' + Key + '"');
            Body.Add('');
            Body.Add(Val); //     #13#10 -  
          end;
        end;
        PostDt.Free;
      end;
      
      L := 0;
      
      //    
      if FULStream <> nil then
      begin
        Body.Add('--' + _prop_Boundary);
        
        if _prop_FNOverride <> '' then
          Val := _prop_FNOverride
        else
          if FdtUploadFile <> '' then
            Val := ExtractFileName(FdtUploadFile)
          else
            Val := DefaultUploadFN;
        
        Key := Val; //     
        if _prop_URLCodePage = 0 then
          Val := CodePage1ToCodePage2(Val, CP_ACP, CP_UTF8);
        Body.Add('Content-Disposition: form-data; name="' + _prop_FileField + '"; filename="' + Val + '"');
        
        if _prop_FileMIME <> '' then
          Val := _prop_FileMIME
        else
          Val := MimeTypeByFileExt(Key);
        Body.Add('Content-Type: ' + Val);
        
        Body.Add(''); //    #13#10
        Body.Add(''); //       -    #13#10
        L := FULStream.Size;
      end;

      FReqBody := Body.Text;
      //   #13#10 (    Boundary      )
      FReqBody := Copy(FReqBody, 1, Length(FReqBody) - 2);
      
      //   \20\20   
      //     \r\n    POSTData  EncType=multipart/form-data
      //   "="  ""  "\r\n"  "\20\20"
      Replace(FReqBody, #20#20, #13#10);
      
      Body.Free;
      FBodyLength := DWord(Length(FReqBody)) + L; //   ,  +  
      
      if FBodyLength > 0 then
      begin
        FBodyLength := FBodyLength + DWord(Length(_prop_Boundary)) + 8; // +    #13#10'--'+_prop_Boundary+'--'#13#10
        FReqHeaders := FReqHeaders +
          'Content-Type: multipart/form-data; boundary=' + _prop_Boundary + #13#10;
          //+'Content-Length: '+UInt2Str(FBodyLength) + #13#10;
      end;
    end
    else // _prop_EncType <> ENC_TYPE_MULTIPART
    begin
      if FdtPostData <> '' then
      begin
        if _prop_EncType = ENC_TYPE_URL_ENC then
        begin
          Replace(FdtPostData, #13#10, '&');
          FReqHeaders := FReqHeaders +
            'Content-Type: application/x-www-form-urlencoded'#13#10;
        end;
        FReqBody := FdtPostData;
        FBodyLength := Length(FReqBody);
        //FReqHeaders := FReqHeaders +
        //  'Content-Length: '+UInt2Str(FBodyLength) + #13#10;
      end;
    end;
    
    FReqHeaders := FReqHeaders +
      'Content-Length: '+UInt2Str(FBodyLength) + #13#10;
  
  end; // FNeedBody
  
  if FdtHeaders <> '' then
    FReqHeaders := FReqHeaders + FdtHeaders;
end;


function ThiHTTPClient.StartRequest: DWord;
var
  N: Cardinal;
  WStrP1, WStrP2: PWideChar;
begin
  //Result := ERROR_OPERATION_ABORTED;
  
  FURLParser.URI := FdtURL;
  if FURLParser.Protocol <> 'https' then FURLParser.Protocol := 'http';
  
  
  N := WINHTTP_ACCESS_TYPE_NO_PROXY;
  WStrP1 := nil;
  WStrP2 := nil;
  
  
  if _prop_UseSystemProxy = 0 then
  begin
    if FdtProxy <> '' then
    begin
      N := WINHTTP_ACCESS_TYPE_NAMED_PROXY; 
      WStrP1 := PWideChar(WideString(FdtProxy));
      WStrP2 := PWideChar(WideString('<local>'));
    end;
  end
  else
    N := WINHTTP_ACCESS_TYPE_DEFAULT_PROXY;
  
  // --------- WinHttpOpen --------- //
  
  FHSession := WinHttpOpen(
    PWideChar(WideString(_prop_UserAgent)),
    N,
    WStrP1,
    WStrP2,
    0
  );

  if FHSession = nil then
  begin
    Result := GetLastError;
    Exit;
  end;
  
  //SetHTTPOptionDWord(FHSession, WINHTTP_OPTION_CONNECT_TIMEOUT, _prop_ConnectTimeout);
  //SetHTTPOptionDWord(FHSession, WINHTTP_OPTION_SEND_TIMEOUT, _prop_SendTimeout);
  //SetHTTPOptionDWord(FHSession, WINHTTP_OPTION_RECEIVE_TIMEOUT, _prop_RecvTimeout);
  WinHttpSetTimeouts(FHSession, 0, _prop_ConnectTimeout, _prop_SendTimeout, _prop_RecvTimeout);
  
  if _prop_URLCodePage <> 0 then
    SetHTTPOptionDWord(FHSession, WINHTTP_OPTION_CODEPAGE, CP_ACP);
  
  // -------------------------------- //


  //       -  
  if FURLParser.Port = '' then
  begin
    if FURLParser.Protocol = 'https' then N := 443 else N := 80;
  end
  else
    N := Str2Int(FURLParser.Port);


  // --------- WinHttpConnect --------- //
  
  FHConnect := WinHttpConnect(
    FHSession,
    PWideChar(WideString(FURLParser.Host)),
    N,
    0
  );

  if FHConnect = nil then
  begin
    Result := GetLastError;
    Exit;
  end;
  
  // ----------------------------------- //

  
  //if FdtReferer <> '' then WStrP1 := PWideChar(WideString(FdtReferer)) else WStrP1 := nil;
  
  
  N := WINHTTP_FLAG_REFRESH;
  if FURLParser.Protocol = 'https' then N := N or WINHTTP_FLAG_SECURE;
  
  
  // --------- WinHttpOpenRequest --------- //
  
  FHRequest := WinHttpOpenRequest(
    FHConnect,
    PWideChar(FdtMethod), //pwszVerb
    PWideChar(WideString(FURLParser.GetFullURI([ofParams]))), //pwszObjectName
    'HTTP/1.1', //pwszVersion
    //WStrP1, //pwszReferrer
    Pointer(WideString(FdtReferer)), //pwszReferrer
    WINHTTP_DEFAULT_ACCEPT_TYPES, //ppwszAcceptTypes //
    N
  );

  if FHRequest = nil then
  begin
    Result := GetLastError;
    Exit;
  end;

  // ----------------------------------- //
  
  // Callback 
  WinHttpSetStatusCallback(FHRequest, @StatusCallback, CallbackFlags, 0);
  
  //  
  if _prop_AutoRedirect = 0 then
    SetHTTPOptionDWord(FHRequest, WINHTTP_OPTION_DISABLE_FEATURE, WINHTTP_DISABLE_REDIRECTS);
  
  if _prop_SSLProtocols <> 0 then
    SetHTTPOptionDWord(FHRequest, WINHTTP_OPTION_SECURE_PROTOCOLS, WINHTTP_FLAG_SECURE_PROTOCOL_ALL);
  
  N := 
    (_prop_IgnoreInvalidCN * SECURITY_FLAG_IGNORE_CERT_CN_INVALID) or
    (_prop_IgnoreInvalidDate * SECURITY_FLAG_IGNORE_CERT_DATE_INVALID) or
    (_prop_IgnoreUnknownCA * SECURITY_FLAG_IGNORE_UNKNOWN_CA) or
    (_prop_IgnoreWrongUsage * SECURITY_FLAG_IGNORE_CERT_WRONG_USAGE) or
    (_prop_IgnoreWeakSignature * SECURITY_FLAG_IGNORE_WEAK_SIGNATURE);
  
  //  :
  {N := 0;
  if _prop_IgnoreInvalidCN <> 0 then N := SECURITY_FLAG_IGNORE_CERT_CN_INVALID;
  if _prop_IgnoreInvalidDate <> 0 then N := N or SECURITY_FLAG_IGNORE_CERT_DATE_INVALID;
  if _prop_IgnoreUnknownCA <> 0 then N := N or SECURITY_FLAG_IGNORE_UNKNOWN_CA;
  if _prop_IgnoreWrongUsage <> 0 then N := N or SECURITY_FLAG_IGNORE_CERT_WRONG_USAGE;
  if _prop_IgnoreWeakSignature <> 0 then N := N or SECURITY_FLAG_IGNORE_WEAK_SIGNATURE;}
  
  if N <> 0 then
    SetHTTPOptionDWord(FHRequest, WINHTTP_OPTION_SECURITY_FLAGS, N);
  
  Result := NO_ERROR;
end;


function ThiHTTPClient.SendRequest: DWord;
var
  ProxyAuthSended, CanSendData, DataSended: Boolean;
  CurrUrl, LastAuthUrl: string;
  
label
  send_again;
begin
  Result := ERROR_OPERATION_ABORTED;
  
  FRedirected := False;
  DataSended := False;
  ProxyAuthSended := False;
  LastAuthUrl := '';
  
  CanSendData := (_prop_AuthCheck = AUTH_CHECK_DISABLED);
  
  // ,    
  CurrUrl := GetHTTPOptionString(FHRequest, WINHTTP_OPTION_URL);

  
send_again:

  // ------    ------ //
  
  //   
  if _prop_CookieManager <> nil then
    AddHeadersToRequest(FHRequest, _prop_CookieManager.GetCookies(CurrUrl));

  //  
  AddHeadersToRequest(FHRequest, FReqHeaders);

  // ----------------------------------- //


  
  // --------- WinHttpSendRequest --------- //
  
  if not WinHttpSendRequest(
           FHRequest,
           nil, //pwszHeaders
           0, //dwHeadersLength
           nil, //lpOptional
           0, //dwOptionalLength
           FBodyLength, //dwTotalLength
           Self
        )
  then
  begin
    Result := GetLastError;
    Exit;
  end;

  // ----------------------------------- //
  
  
  if FAbort then Exit;

  
  // ---------    POST --------- //
  
  if CanSendData and FNeedBody and (not FRedirected) then //     
  begin
    Result := SendBody();
    if Result <> NO_ERROR then Exit;
  end;

  // ----------------------------------- //

  // --------- WinHttpReceiveResponse --------- //
  
  if not WinHttpReceiveResponse(FHRequest, nil) then
  begin
    Result := GetLastError;
    Exit;
  end;

  // ----------------------------------- //
 
  // ,     
  CurrUrl := GetHTTPOptionString(FHRequest, WINHTTP_OPTION_URL);
  
  FStatusHeader := Str2Int(ReadResponseHeader(FHRequest, WINHTTP_QUERY_STATUS_CODE));
  FContLenHeader := Trunc(Str2Double(ReadResponseHeader(FHRequest, WINHTTP_QUERY_CONTENT_LENGTH)));
  
  //            .
  //   ContentLength,         ( onDownloadProgress).
  // TODO:    ,       .
  if (_prop_ResumeDL <> 0) and (FStatusHeader = 206 {Partial Content}) then
  begin
    Inc(FContLenHeader, StreamGetSize(FDLStream));
  end;
  
  FContTypeHeader := ReadResponseHeader(FHRequest, WINHTTP_QUERY_CONTENT_TYPE);
  FResponseHeaders := ReadResponseHeader(FHRequest, HTTP_QUERY_RAW_HEADERS_CRLF);

  if _prop_CookieManager <> nil then
  begin
    _prop_CookieManager.AddCookies(CurrUrl, ReadResponseCookies(FHRequest));
  end;
  
  // Proxy 
  if (FStatusHeader = 407) and (not ProxyAuthSended) then // TODO:   -    
  begin
    ProxyAuthSended := True;
    SetAuthCredentials(FHRequest, 
      FdtProxyUser, 
      FdtProxyPass, 
      WINHTTP_AUTH_TARGET_PROXY,
      ChooseAuthScheme(FHRequest)
    );
    CanSendData := _prop_AuthCheck <> AUTH_CHECK_ALL;
    DataSended := CanSendData;
    goto send_again;
  end;

  // HTTP 
  if (FStatusHeader = 401) and (LastAuthUrl <> CurrUrl) then //       URL      -  /
  begin
    LastAuthUrl := CurrUrl;

    SetAuthCredentials(FHRequest, 
      FdtHTTPUser, 
      FdtHTTPPass, 
      WINHTTP_AUTH_TARGET_SERVER,
      ChooseAuthScheme(FHRequest)
    );
    CanSendData := True;
    DataSended := True;
    goto send_again;
  end;
  
  if (_prop_AuthCheck <> AUTH_CHECK_DISABLED) and
     (FStatusHeader = 200) and
     (not FRedirected) and
     (not DataSended)
  then
  begin
    CanSendData := True;
    DataSended := True;
    goto send_again;
  end;
  
  Result := NO_ERROR;
end;


function ThiHTTPClient.SendBody: DWord;
var
  L, I, R, Sz: DWord;
  Buf: array of Byte;
  FinalBound: string;
const
  READ_BUFFER_SIZE = 8*1024;
begin
  Result := ERROR_OPERATION_ABORTED;
  
  if {(FBodyLength = 0) or} (FReqBody = '') then Exit; // FReqBody <> ''    ,   POSTData
  
  I := 0;
  Sz := Length(FReqBody);
  while I < Sz do
  begin
    R := Sz - I;
    if R > READ_BUFFER_SIZE then R := READ_BUFFER_SIZE;
    
    if not WinHttpWriteData(FHRequest, @FReqBody[I + 1], R, L) then
    begin
      Result := GetLastError;
      Exit;
    end;
    
    if FAbort then Exit;
    
    Inc(I, R);
  end;
  
  if FULStream <> nil then
  begin
    FULStream.Position := 0;
    SetLength(Buf, READ_BUFFER_SIZE);
    Sz := FULStream.Size;
    I := 0;
    while I < Sz do
    begin
      R := FULStream.Read(Buf[0], READ_BUFFER_SIZE);
      if R > 0 then
      begin
        if not WinHttpWriteData(FHRequest, Pointer(Buf), R, L) then
        begin
          Result := GetLastError;
          Exit;
        end;
        
        if FAbort then Exit;
        
        Inc(I, R);
        _hi_OnEvent(_event_onUploadProgress, Integer(I));
      end;
      if R < READ_BUFFER_SIZE then Break;
    end;
  end;
  
  if FAbort then Exit;
  
  if _prop_EncType = ENC_TYPE_MULTIPART then
  begin
    FinalBound := #13#10'--' + _prop_Boundary + '--'#13#10;
    WinHttpWriteData(FHRequest, Pointer(FinalBound), Length(FinalBound), L);
  end;
  
  Result := NO_ERROR;
end;


function ThiHTTPClient.SaveResponse: DWord;
var
  Buf: array of Byte;
  BytesRead: DWord;
  Total: Double;
const
  BUFFER_SIZE = 10*1024;
begin
  if FDLStream = nil then
  begin
    Result := NO_ERROR;
    Exit;
  end;
  
  Result := ERROR_OPERATION_ABORTED;
  
  //          .
  //     .
  if _prop_ResumeDL <> 0 then
  begin
    // 416 - Requested Range Not Satisfiable
    //         .
    
    //        200 (OK), 206 (Partial Content).
    //         ,    .
    if not (FStatusHeader in [200, 206]) then
    begin
      Result := NO_ERROR; //   ,       
      Exit;
    end;
    
    Total := StreamGetSize(FDLStream);
  end
  else
    Total := 0;
  
  //     Memory-
  if (FDLStream.Memory <> nil) and (FContLenHeader > BUFFER_SIZE) then
  begin
    if _prop_ResumeDL <> 0 then
      FDLStream.Capacity := FContLenHeader - FDLStream.Size // FContLenHeader   
    else
      FDLStream.Capacity := FContLenHeader;
  end;
  
  SetLength(Buf, BUFFER_SIZE);
  
  while not FAbort do
  begin
    if not WinHttpReadData(FHRequest, Pointer(Buf), BUFFER_SIZE, BytesRead) then
    begin
      Result := GetLastError;
      Exit;
    end;
    
    if FAbort then Exit;
    
    if BytesRead = 0 then //  
    begin
      Result := NO_ERROR;
      SetEndOfFile(FDLStream.Handle);
      Exit;
    end;
    
    if FDLStream.Write(Buf[0], BytesRead) < BytesRead then
    begin
      //    
      Result := GetLastError;
      Exit;
    end;
    
    if FAbort then Exit;
    
    Total := Total + BytesRead;
    _hi_OnEvent(_event_onDownloadProgress, Total);
  end;
  
end;


procedure ThiHTTPClient.Cleanup(IsFailed: Boolean);
var
  NeedDel: Boolean;
begin
  FRedirected := False;
  FNeedBody := False;
  FBodyLength := 0;
  FReqBody := '';
  FReqHeaders := '';

  if FReleaseDLStream then
  begin
    //      ,
    //     
    NeedDel := IsFailed and (FdtFileName <> '') and (FDLStream <> nil) and (FDLStream.Size = 0);
    
    FDLStream.Free;
    
    if NeedDel then DeleteFile(PChar(FdtFileName));
  end;

  if FReleaseULStream then FULStream.Free;

  FDLStream := nil;
  FULStream := nil;

  FReleaseDLStream := False;
  FReleaseULStream := False;

  FdtURL := '';
  FdtFileName := '';
  FdtHeaders := '';
  FdtHTTPUser := '';
  FdtHTTPPass := '';
  FdtMethod := '';
  FdtPostData := '';
  FdtProxy := '';
  FdtProxyUser := '';
  FdtProxyPass := '';
  FdtReferer := '';
  FdtUploadFile := '';
  FdtUploadStream := nil;

  CancelRequest;
end;

procedure ThiHTTPClient.CancelRequest;
begin
  FAbort := True;

  if FHRequest <> nil then WinHttpCloseHandle(FHRequest);
  if FHConnect <> nil then WinHttpCloseHandle(FHConnect);
  if FHSession <> nil then WinHttpCloseHandle(FHSession);
  
  FHRequest := nil;
  FHConnect := nil;
  FHSession := nil;
end;





procedure ThiHTTPClient._work_doAbort(var _Data: TData; Index: Word);
begin
  CancelRequest;
end;

procedure ThiHTTPClient._work_doBoundary(var _Data: TData; Index: Word);
begin
  _prop_Boundary := ToString(_Data);
end;

procedure ThiHTTPClient._work_doCustomMethod(var _Data: TData; Index: Word);
begin
  _prop_CustomMethod := ToString(_Data);
end;

procedure ThiHTTPClient._work_doEncType(var _Data: TData; Index: Word);
begin
  _prop_EncType := ToInteger(_Data);
end;

procedure ThiHTTPClient._work_doExecute(var _Data: TData; Index: Word);
begin
  InternalDownload(_Data, PROMPT_FOR_SAVE);
end;

procedure ThiHTTPClient._work_doFileField(var _Data: TData; Index: Word);
begin
  _prop_FileField := ToString(_Data);
end;

procedure ThiHTTPClient._work_doFileMIME(var _Data: TData; Index: Word);
begin
  _prop_FileMIME := ToString(_Data);
end;

procedure ThiHTTPClient._work_doFNOverride(var _Data: TData; Index: Word);
begin
  _prop_FNOverride := ToString(_Data);
end;

procedure ThiHTTPClient._work_doLoadFile(var _Data: TData; Index: Word);
begin
  InternalDownload(_Data, SAVE_TO_FILE);
end;

procedure ThiHTTPClient._work_doLoadStream(var _Data: TData; Index: Word);
begin
  InternalDownload(_Data, SAVE_TO_STREAM);
end;

procedure ThiHTTPClient._work_doLoadString(var _Data: TData; Index: Word);
begin
  InternalDownload(_Data, SAVE_TO_STRING);
end;

procedure ThiHTTPClient._work_doResumeDL(var _Data: TData; Index: Word);
begin
  _prop_ResumeDL := ToInteger(_Data);
end;

procedure ThiHTTPClient._work_doUseSystemProxy(var _Data: TData; Index: Word);
begin
  _prop_UseSystemProxy := ToInteger(_Data);
end;



procedure ThiHTTPClient._var_Status(var _Data: TData; Index: Word);
begin
  dtInteger(_Data, FStatusHeader);
end;

procedure ThiHTTPClient._var_ContentLength(var _Data: TData; Index: Word);
begin
  dtReal(_Data, FContLenHeader);
end;

procedure ThiHTTPClient._var_ContentType(var _Data: TData; Index: Word);
begin
  dtString(_Data, FContTypeHeader);
end;

procedure ThiHTTPClient._var_CurrentURL(var _Data: TData; Index: Word);
begin
  dtString(_Data, GetHTTPOptionString(FHRequest, WINHTTP_OPTION_URL));
end;

procedure ThiHTTPClient._var_RawHeaders(var _Data: TData; Index: Word);
begin
  dtString(_Data, FResponseHeaders);
end;

procedure ThiHTTPClient._var_StatusText(var _Data: TData; Index: Word);
begin
  dtString(_Data, ReadResponseHeader(FHRequest, HTTP_QUERY_STATUS_TEXT));
end;


end.
